/**
 * Copyright (C) 2010-2011 J.W.Marsden
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/
package jsonij.json;

import java.io.IOException;
import java.io.StringReader;
import jsonij.parser.ParserException;
import jsonij.parser.Position;
import jsonij.parser.ReaderParser;

/**
 * java.lang.String implementation of a JSONReader. 
 * @author J.W.Marsden
 */
public class StringJSONReader extends JSONReader {

    final protected String jsonString;
    protected int index;
    protected int length;
    protected StringJSONStringReader stringReader;

    public StringJSONReader(String jsonString) {
        if (jsonString == null) {
            throw new NullPointerException("Null String Exception.");
        }
        this.jsonString = jsonString;
        length = jsonString.length();
        index = 0;
    }

    @Override
    public int readNext() throws IOException, ParserException {
        int r = -1;
        while (index < length) {
            r = jsonString.charAt(index++);
            //position.tickReadCount();
            position.movePosition();
            if (ConstantUtility.isReturn(r)) {
                handleNewLine();
                r = jsonString.charAt(index++);
                if (index < length && ConstantUtility.isNewLine(r)) {
                    r = jsonString.charAt(index++);
                }
            } else if (ConstantUtility.isNewLine(r)) {
                handleNewLine();
                r = jsonString.charAt(index++);
                if (index < length && ConstantUtility.isReturn(r)) {
                    r = jsonString.charAt(index++);
                }
                
            }
            if (!ConstantUtility.isWhiteSpace(r)) {
                break;
            }
        }
        return r;
    }

    public ReaderParser getStringReader() {
        if (stringReader == null) {
            stringReader = new StringJSONStringReader();
        } else if (stringReader.isActive()) {
            // TODO: Do we want an exception here?
        }
        stringReader.setActive(true);
        return stringReader;
    }

    protected class StringJSONStringReader implements ReaderParser {

        protected boolean active;

        public StringJSONStringReader() {
            active = true;
        }

        /**
         * @return the active
         */
        public boolean isActive() {
            return active;
        }

        /**
         * @param active the active to set
         */
        protected void setActive(boolean active) {
            this.active = active;
        }

        public int peek() throws IOException {
            if (stringReader == null || !active) {
                // TODO: Dead Reader!
            }
            if (!hasPeeked) {
                peekValue = readNext();
                hasPeeked = true;
            }
            return peekValue;
        }

        public int read() throws IOException {
            if (stringReader == null || !active) {
                // TODO: Dead Reader!
            }
            if (hasPeeked) {
                hasPeeked = false;
                return peekValue;
            }
            return readNext();
        }

        public void close() {
            active = false;
        }

        protected int readNext() throws IOException {
            if (jsonString == null) {
                throw new NullPointerException("Null String Exception.");
            }
            int r = -1;
            while (index < length) {
                r = jsonString.charAt(index++);
                //position.tickReadCount();
                position.movePosition();
                break;
            }
            return r;
        }

        public Position getPosition() {
            return StringJSONReader.this.getPosition();
        }

        public boolean isHasPeeked() {
            return StringJSONReader.this.isHasPeeked();
        }

        public boolean hasPeeked() {
            return StringJSONReader.this.hasPeeked();
        }

        public void setHasPeeked(boolean hasPeeked) {
            StringJSONReader.this.setHasPeeked(hasPeeked);
        }

        public int getLineNumber() {
            return StringJSONReader.this.getLineNumber();
        }

        public int getPositionNumber() {
            return StringJSONReader.this.getPositionNumber();
        }
    }
}
